"""
This module provides the EntropyBasedMethod class.

Classes:
    EntropyBasedMethod: EntropyBasedMethod abstract class.
"""

from abc import ABC, abstractmethod
from copy import deepcopy
from causalflow.preprocessing.subsampling_methods.moving_window import MovingWindow

    
class EntropyBasedMethod(ABC):
    """EntropyBasedMethod abstract class."""
    
    def __init__(self, threshold):
        """
        Class constructor.

        Args:
            threshold (float): entropy threshold.
        """
        self.windows = list()
        self.segments = list()
        self.threshold = threshold


    def create_rounded_copy(self):
        """
        Create deepcopy of the dataframe but with rounded values.

        Returns:
            (pd.DataFrame): rounded dataframe.
        """
        de = deepcopy(self.df)
        de = de.round(1)
        return de


    def __normalization(self):
        """Normalize entropy for each moving window."""
        max_e = max([mw.entropy for mw in self.windows])
        for mw in self.windows:
            mw.entropy = mw.entropy / max_e


    def moving_window_analysis(self):
        """Compute dataframe entropy on moving windows."""
        de = self.create_rounded_copy()

        for ll, rl in self.segments:
            # Create moving window
            mw_df = de.values[ll: rl]

            # Build a Moving Window
            mw = MovingWindow(mw_df)

            # Compute entropy
            mw.get_entropy()

            # Compute optimal number of samples
            mw.optimal_sampling(self.threshold)

            # Collect result in a list
            self.windows.append(mw)

        # Entropy normalization
        self.__normalization()


    def extract_indexes(self):
        """Extract a list of indexes corresponding to the samples selected by the subsampling procedure."""
        _sample_index_list = list()
        for i, mw in enumerate(self.windows):
            sum_ws = sum([wind.T for wind in self.windows[:i]])
            sample_index = [si + sum_ws for si in mw.opt_samples_index]
            _sample_index_list += sample_index
        return _sample_index_list


    @abstractmethod
    def dataset_segmentation(self):
        """Abstract method."""
        pass


"""This module provides methods for computing various metrics."""

from copy import deepcopy

def fully_connected_dag(features, min_lag, max_lag, alsoOrient = False) -> dict:
    """
    Build a fully connected DAG.

    Args:
        features (list): variable list
        min_lag (int): min time lag.
        max_lag (int): max time lag.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        dict: fully connected dag
    """
    if not alsoOrient:
        g = {f: list() for f in features}
        for t in g:
            for s in g:
                for l in range(min_lag, max_lag + 1):
                    if s == t and l == 0: continue 
                    g[t].append((s, -abs(l)))
    else:
        g = {f: {} for f in features}
        for t in g:
            for s in g:
                for l in range(min_lag, max_lag + 1):
                    if s == t and l == 0: continue 
                    g[t][(s, -abs(l))] = ['-->', 'o->', '<->', 'o-o']
    return g


def get_TP(gt, cm, alsoOrient = False) -> int:
    """
    Compute true positive number: edge present in the causal model and present in the groundtruth.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        int: true positive.
    """
    counter = 0
    if not alsoOrient:
        for t in cm.keys():
            for s in cm[t]:
                if s in gt[t]: counter += 1
    else:
        for t in cm.keys():
            for s in cm[t]:
                if s in gt[t]:
                    if gt[t][s][0] == cm[t][s][0]: counter += 1
                    if gt[t][s][1] == cm[t][s][1]: counter += 1
                    if gt[t][s][2] == cm[t][s][2]: counter += 1
    return counter

 
def get_TN(gt, min_lag, max_lag, cm, alsoOrient = False) -> int:
    """
    Compute true negative number: edge absent in the groundtruth and absent in the causal model.
    
    Args:
        gt (dict): ground-truth SCM.
        min_lag (int): min time lag.
        max_lag (int): max time lag.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.
        
    Returns:
        int: true negative.
    """
    fullg = fully_connected_dag(list(gt.keys()), min_lag, max_lag, alsoOrient)
    counter = 0
    if not alsoOrient:
        gt_TN = deepcopy(fullg)
        
        # Build the True Negative graph [complementary graph of the ground-truth]
        for t in fullg:
            for s in fullg[t]:
                if s in gt[t]:
                    gt_TN[t].remove(s)
                    
        for t in gt_TN.keys():
            for s in gt_TN[t]:
                if s not in cm[t]: counter += 1
    else:
        gt_TN = deepcopy(fullg)
        
        # Build the True Negative graph [complementary graph of the ground-truth]
        for t in fullg:
            for s in fullg[t]:
                for edge in fullg[t][s]:
                    if s not in gt[t]: continue
                    if edge == gt[t][s]:
                        gt_TN[t][s].remove(edge)
                if len(fullg[t][s]) == 0: del gt_TN[t][s]
                    
        for t in gt_TN.keys():
            for s in gt_TN[t]:
                if s not in cm[t]: 
                    counter += len(gt_TN[t][s])
                else:
                    for edge in gt_TN[t][s]:
                        if edge != cm[t][s]: counter += 1
        
    return counter
    
     
def get_FP(gt, cm, alsoOrient = False) -> int:
    """
    Compute false positive number: edge present in the causal model but absent in the groundtruth.
    
    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.
        
    Returns:
        int: false positive.
    """
    counter = 0
    if not alsoOrient:
        for t in cm.keys():
            for s in cm[t]:
                if s not in gt[t]: counter += 1
    else:
        for t in cm.keys():
            for s in cm[t]:
                if s not in gt[t]: counter += 3
                else:
                    if gt[t][s][0] != cm[t][s][0]: counter += 1
                    if gt[t][s][1] != cm[t][s][1]: counter += 1
                    if gt[t][s][2] != cm[t][s][2]: counter += 1
    return counter


def get_FN(gt, cm, alsoOrient = False) -> int:
    """
    Compute false negative number: edge present in the groundtruth but absent in the causal model.
        
    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.
        
    Returns:
        int: false negative.
    """
    counter = 0
    if not alsoOrient:
        for t in gt.keys():
            for s in gt[t]:
                if s not in cm[t]: counter += 1
    else:
        for t in gt.keys():
            for s in gt[t]:
                if s not in cm[t]: counter += 3
                else:
                    if gt[t][s][0] != cm[t][s][0]: counter += 1
                    if gt[t][s][1] != cm[t][s][1]: counter += 1
                    if gt[t][s][2] != cm[t][s][2]: counter += 1
    return counter
    
    
def shd(gt, cm, alsoOrient = False) -> int:
    """
    Compute Structural Hamming Distance between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        int: shd.
    """
    fn = get_FN(gt, cm, alsoOrient)
    fp = get_FP(gt, cm, alsoOrient)
    return fn + fp


def precision(gt, cm, alsoOrient = False) -> float:
    """
    Compute Precision between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: precision.
    """
    tp = get_TP(gt, cm, alsoOrient)
    fp = get_FP(gt, cm, alsoOrient)
    if tp + fp == 0: return 0
    return tp/(tp + fp)

        
def recall(gt, cm, alsoOrient = False) -> float:
    """
    Compute Recall between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: recall.
    """
    tp = get_TP(gt, cm, alsoOrient)
    fn = get_FN(gt, cm, alsoOrient)
    if tp + fn == 0: return 0
    return tp/(tp + fn)


def f1_score(gt, cm, alsoOrient = False) -> float:
    """
    Compute F1-score between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: F1-score.
    """
    p = precision(gt, cm, alsoOrient)
    r = recall(gt, cm, alsoOrient)
    if p + r == 0: return 0
    return (2 * p * r) / (p + r)
    
     
def FPR(gt, min_lag, max_lag, cm, alsoOrient = False) -> float:
    """
    Compute False Positve Rate between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        min_lag (int): min time lag.
        max_lag (int): max time lag.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: false positive rate.
    """
    fp = get_FP(gt, cm, alsoOrient)
    tn = get_TN(gt, min_lag, max_lag, cm, alsoOrient)
    if tn + fp == 0: return 0
    return fp / (tn + fp)
    
    
def TPR(gt, cm, alsoOrient = False) -> float:
    """
    Compute True Positive Rate between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: true positive rate.
    """
    tp = get_TP(gt, cm, alsoOrient)
    fn = get_FN(gt, cm, alsoOrient)
    if tp + fn == 0: return 0
    return tp / (tp + fn)


def TNR(gt, min_lag, max_lag, cm, alsoOrient = False) -> float:
    """
    Compute True Negative Rate between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        min_lag (int): min time lag.
        max_lag (int): max time lag.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: true negative rate.
    """
    tn = get_TN(gt, min_lag, max_lag, cm, alsoOrient)
    fp = get_FP(gt, cm, alsoOrient)
    if tn + fp == 0: return 0
    return tn / (tn + fp)


def FNR(gt, cm, alsoOrient = False) -> float:
    """
    Compute False Negative Rate between ground-truth causal graph and the estimated one.

    Args:
        gt (dict): ground-truth SCM.
        cm (dict): estimated SCM.
        alsoOrient (bool, optional): Flag to compute the metric considering also the orientation of the link and not only its presence. Default False.

    Returns:
        float: false negative rate.
    """
    fn = get_FN(gt, cm, alsoOrient)
    tp = get_TP(gt, cm, alsoOrient)
    if tp + fn == 0: return 0
    return fn / (tp + fn)


#!/usr/bin/env python
# coding: utf-8
"""
Functions for parsing various graph formats and converting them into a node list, edge list, and edge-to-weight mapping.
"""

import warnings
import numpy as np

from functools import wraps

from ._utils import (
    _get_unique_nodes,
    _save_cast_float_to_int,
)


def _handle_multigraphs(parser):
    """Raise a warning if the given graph appears to be a multigraph, and remove duplicate edges."""
    @wraps(parser)
    def wrapped_parser(graph, *args, **kwargs):
        nodes, edges, edge_weight = parser(graph, *args, **kwargs)

        unique_edges = list(set(edges))
        if len(unique_edges) < len(edges):
            msg = "The given graph data structure appears to be a multi-graph,"
            msg += " however, this netgraph class does not properly support multi-graphs."
            msg += " Instead, the multi-graph is converted into a weighted graph,"
            msg += " in which duplicate edges are merged into one."
            msg += " The corresponding edge weight is set to the number of duplicate edges;"
            msg += " existing edge weights are discarded."
            msg += " Use the `MultiGraph` class and subclasses to visualize all edges and their edge weights properly."
            warnings.warn(msg)

            edge_weight = dict()
            for edge in edges:
                edge_weight[edge] = edge_weight.get(edge, 0) + 1

            return nodes, unique_edges, edge_weight

        return nodes, edges, edge_weight

    return wrapped_parser


def _parse_edge_list(edges):
    """Ensures that the type of edges is a list, and each edge is a 2-tuple."""
    # Edge list may be an array, or a list of lists. We want a list of tuples.
    return [(source, target) for (source, target) in edges]


@_handle_multigraphs
def _parse_sparse_matrix_format(adjacency):
    """Parse graphs given in a sparse format, i.e. edge lists or sparse matrix representations."""
    rows, columns = np.array(adjacency).shape

    if columns == 2:
        edges = _parse_edge_list(adjacency)
        nodes = _get_unique_nodes(edges)
        return nodes, edges, None

    elif columns == 3:
        edge_weight = {(source, target) : weight for (source, target, weight) in adjacency}
        edges = list(edge_weight.keys())
        nodes = _get_unique_nodes(edges)

        # In a sparse adjacency format with integer nodes and float weights,
        # the type of nodes is promoted to the same type as weights.
        # If all nodes can safely be demoted to ints, then we probably want to do that.
        save = True
        for node in nodes:
            if not isinstance(_save_cast_float_to_int(node), int):
                save = False
                break
        if save:
            nodes = [_save_cast_float_to_int(node) for node in nodes]
            edges = [(_save_cast_float_to_int(source), _save_cast_float_to_int(target)) for (source, target) in edges]
            edge_weight = {(_save_cast_float_to_int(source), _save_cast_float_to_int(target)) : weight for (source, target), weight in edge_weight.items()}

        if len(set(edge_weight.values())) > 1:
            return nodes, edges, edge_weight
        else:
            return nodes, edges, None

    else:
        msg = "Graph specification in sparse matrix format needs to consist of an iterable of tuples of length 2 or 3."
        msg += "Got iterable of tuples of length {}.".format(columns)
        raise ValueError(msg)


def _parse_adjacency_matrix(adjacency):
    """Parse graphs given in adjacency matrix format, i.e. a full-rank matrix."""
    sources, targets = np.where(adjacency)
    edges = list(zip(sources.tolist(), targets.tolist()))
    nodes = list(range(adjacency.shape[0]))
    edge_weights = {(source, target): adjacency[source, target] for (source, target) in edges}

    if len(set(list(edge_weights.values()))) == 1:
        return nodes, edges, None
    else:
        return nodes, edges, edge_weights


def _parse_nparray(graph):
    rows, columns = graph.shape
    if columns in (2, 3):
        return _parse_sparse_matrix_format(graph)
    elif rows == columns:
        return _parse_adjacency_matrix(graph)
    else:
        msg = "Could not interpret input graph."
        msg += "\nIf a graph is specified as a numpy array, it has to have one of the following shapes:"
        msg += "\n\t-(E, 2) or (E, 3), where E is the number of edges"
        msg += "\n\t-(V, V), where V is the number of nodes (i.e. full rank)"
        msg += f"\nHowever, the given graph had shape {graph.shape}."
        raise ValueError(msg)


@_handle_multigraphs
def _parse_networkx_graph(graph, weight_attribute="weight"):
    """Parse graphs represented as networkx.Graph or related objects."""
    edges = list(graph.edges)
    nodes = list(graph.nodes)
    try:
        edge_weights = {edge : graph.get_edge_data(*edge)[weight_attribute] for edge in edges}
    except KeyError: # no weights
        edge_weights = None
    return nodes, edges, edge_weights


@_handle_multigraphs
def _parse_igraph_graph(graph, weight_attribute="weight"):
    """Parse graphs given as igraph.Graph or related objects."""
    edges = [(edge.source, edge.target) for edge in graph.es()]
    nodes = graph.vs.indices
    if graph.is_weighted():
        edge_weights = {(edge.source, edge.target) : edge[weight_attribute] for edge in graph.es()}
    else:
        edge_weights = None
    return nodes, edges, edge_weights


@_handle_multigraphs
def _parse_graph_tool_graph(graph):
    """Parse graphs given as graph_tool.Graph."""
    nodes = graph.get_vertices().tolist()
    edges = [tuple(edge) for edge in graph.get_edges()]
    # In graph-tool, edge weights are in separate data structure called an edge property map.
    edge_weights = None
    return nodes, edges, edge_weights


def _is_listlike(graph):
    return isinstance(graph, (list, tuple, set))

def _is_nparray(graph):
    return isinstance(graph, np.ndarray)

def _is_networkx(graph):
    import networkx
    return isinstance(graph, networkx.Graph)

def _is_igraph(graph):
    import igraph
    return isinstance(graph, igraph.Graph)

def _is_graph_tool(graph):
    import graph_tool
    return isinstance(graph, graph_tool.Graph)


_check_to_parser = {
    _is_listlike   : _parse_sparse_matrix_format,
    _is_nparray    : _parse_nparray,
    _is_networkx   : _parse_networkx_graph,
    _is_igraph     : _parse_igraph_graph,
    _is_graph_tool : _parse_graph_tool_graph,
}


def parse_graph(graph):
    """Parse the given graph format and convert it into a node list, edge list, and edge_weight dictionary.

    Parameters
    ----------
    graph: various formats

        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray (where E is the number of edges).
        - Adjacency matrix:
          Full-rank (V, V) ndarray (where V is the number of nodes/vertices).
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.**

        - networkx.Graph, igraph.Graph, or graph_tool.Graph object

    Returns
    -------
    nodes : list
        List of V unique nodes.
    edges: list of 2-tuples
        List of E edges. Each tuple corresponds to an edge defined by (source node, target node).
    edge_weight: dict edge : float or None
        Dictionary mapping edges to weights. If the graph is unweighted, None is returned.

    """
    for check, parser in _check_to_parser.items():
        try:
            if check(graph):
                return parser(graph)
        except ModuleNotFoundError:
            pass
    else:
        allowed = ['list', 'tuple', 'set', 'networkx.Graph', 'igraph.Graph', 'graph_tool.Graph']
        raise NotImplementedError("Input graph must be one of: {}\nCurrently, type(graph) = {}".format("\n\n\t" + "\n\t".join(allowed), type(graph)))


# --------------------------------------------------------------------------------


def is_order_zero(graph):
    """Determine if a graph is an order zero graph, i.e. a graph with no nodes (and no edges)."""
    for check, parser in _check_to_parser.items():
        try:
            if check(graph):
                nodes, edges, _ = parser(graph)
                if (not nodes) and (not edges):
                    return True
                else:
                    return False
        except ModuleNotFoundError:
            pass
    else:
        allowed = ['list', 'tuple', 'set', 'networkx.Graph', 'igraph.Graph', 'graph_tool.Graph']
        raise NotImplementedError("Input graph must be one of: {}\nCurrently, type(graph) = {}".format("\n\n\t" + "\n\t".join(allowed), type(graph)))


def is_empty(graph):
    """Determine if a graph is an empty graph, i.e. a graph with nodes but no edges."""
    for check, parser in _check_to_parser.items():
        try:
            if check(graph):
                nodes, edges, _ = parser(graph)
                if nodes and (not edges):
                    return True
                else:
                    return False
        except ModuleNotFoundError:
            pass
    else:
        allowed = ['list', 'tuple', 'set', 'networkx.Graph', 'igraph.Graph', 'graph_tool.Graph']
        raise NotImplementedError("Input graph must be one of: {}\nCurrently, type(graph) = {}".format("\n\n\t" + "\n\t".join(allowed), type(graph)))


# --------------------------------------------------------------------------------
# multi-graph parsers

def _parse_multigraph_edge_list(edges):
    """Ensures that the type of edges is a list, and each edge is a 2-tuple."""
    # Edge list may be an array, or a list of lists. We want a list of tuples.
    return [(source, target, eid) for (source, target, eid) in edges]


def _parse_multigraph_sparse_matrix_format(adjacency):
    """Parse graphs given in a sparse format, i.e. edge lists or sparse matrix representations."""
    rows, columns = np.array(adjacency).shape

    if columns == 3:
        edges = _parse_multigraph_edge_list(adjacency)
        nodes = _get_unique_nodes(edges)
        return nodes, edges, None

    elif columns == 4:
        edge_weight = {(source, target, eid) : weight for (source, target, eid, weight) in adjacency}
        edges = list(edge_weight.keys())
        nodes = _get_unique_nodes(edges)

        # In a sparse adjacency format with integer nodes and float weights,
        # the type of nodes is promoted to the same type as weights.
        # If all nodes can safely be demoted to ints, then we probably want to do that.
        save = True
        for node in nodes:
            if not isinstance(_save_cast_float_to_int(node), int):
                save = False
                break
        if save:
            nodes = [_save_cast_float_to_int(node) for node in nodes]
            edges = [(_save_cast_float_to_int(source), _save_cast_float_to_int(target), eid) for (source, target, eid) in edges]
            edge_weight = {(_save_cast_float_to_int(source), _save_cast_float_to_int(target), eid) : weight for (source, target, eid), weight in edge_weight.items()}

        if len(set(edge_weight.values())) > 1:
            return nodes, edges, edge_weight
        else:
            return nodes, edges, None

    else:
        msg = "Multi-graph specification in sparse matrix format needs to consist of an iterable of tuples of length 3 or 4."
        msg += "Got iterable of tuples of length {}.".format(columns)
        raise ValueError(msg)


def _parse_multigraph_adjacency_matrix(adjacency):
    """Parse multi-graphs given in an adjacency matrix format."""
    sources, targets, edge_ids = np.where(adjacency)
    edges = list(zip(sources.tolist(), targets.tolist(), edge_ids.tolist()))
    nodes = list(range(adjacency.shape[0]))
    edge_weights = {(source, target, eid): adjacency[source, target, eid] for (source, target, eid) in edges}

    if len(set(edge_weights.values())) == 1:
        return nodes, edges, None
    else:
        return nodes, edges, edge_weights


def _parse_multigraph_nparray(graph):
    if np.ndim(graph) == 2:
        rows, columns = graph.shape
        if columns in (3, 4):
            return _parse_multigraph_sparse_matrix_format(graph)
    elif np.ndim(graph) == 3:
        rows, columns, layers = graph.shape
        if rows == columns:
            return _parse_multigraph_adjacency_matrix(graph)

    msg = "Could not interpret input graph."
    msg += "\nIf a graph is specified as a numpy array, it has to have one of the following shapes:"
    msg += "\n\t-(E, 3) or (E, 4), where E is the number of edges"
    msg += "\n\t-(V, V, L), where V is the number of nodes and L is the number of layers (i.e. each layer is a full rank matrix)"
    msg += f"\nHowever, the given graph had shape {graph.shape}."
    raise ValueError(msg)


# def _parse_multigraph_networkx_graph(graph, weight_attribute="weight"):
#     """Parse graphs represented as networkx.Graph or related objects."""
#     edges = list(graph.edges)
#     nodes = list(graph.nodes)
#     try:
#         edge_weights = {edge : graph.get_edge_data(*edge)[weight_attribute] \
#                         for edge in edges}
#     except KeyError: # no weights
#         edge_weights = None
#     return nodes, edges, edge_weights
_parse_multigraph_networkx_graph = _parse_networkx_graph.__wrapped__


def _parse_multigraph_igraph_graph(graph, weight_attribute="weight", id_attribute="id"):
    """Parse graphs given as igraph.Graph or related objects."""
    edges = [(edge.source, edge.target, edge[id_attribute]) for edge in graph.es()]
    nodes = graph.vs.indices
    if graph.is_weighted():
        edge_weights = {(edge.source, edge.target, edge[id_attribute]) : edge[weight_attribute] for edge in graph.es()}
    else:
        edge_weights = None
    return nodes, edges, edge_weights


def _parse_multigraph_graph_tool_graph(graph):
    return NotImplementedError("Multi-graph plotting is currently not supported for graph-tool Graph objects.")


def _is_networkx_multigraph(graph):
    import networkx
    return isinstance(graph, networkx.MultiGraph)


_check_to_multigraph_parser = {
    _is_listlike            : _parse_multigraph_sparse_matrix_format,
    _is_nparray             : _parse_multigraph_nparray,
    _is_networkx_multigraph : _parse_multigraph_networkx_graph,
    _is_igraph              : _parse_multigraph_igraph_graph,
    _is_graph_tool          : _parse_multigraph_graph_tool_graph,
}


def parse_multigraph(graph):
    """Parse the given multi-graph format and convert it into a node list, edge list, and edge_weight dictionary.

    Parameters
    ----------
    graph: various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source node ID, target node ID, edge key) or
          (source node ID, target node ID, edge key, weight) tuples,
          or equivalent (E, 3) or (E, 4) ndarray (where E is the number of edges).
        - Adjacency matrix:
          A (V, V, L) ndarray (where V is the number of nodes/vertices, and L is the number of layers).
          The absence of a connection is indicated by a zero.
        - networkx.MultiGraph or igraph.Graph object

    Returns
    -------
    nodes : list
        List of V unique nodes.
    edges: list of 2-tuples
        List of E edges. Each tuple corresponds to an edge defined by (source node, target node).
    edge_weight: dict edge : float or None
        Dictionary mapping edges to weights. If the graph is unweighted, None is returned.

    """
    for check, parser in _check_to_multigraph_parser.items():
        try:
            if check(graph):
                return parser(graph)
        except ModuleNotFoundError:
            pass
    else:
        allowed = ['list', 'tuple', 'set', 'networkx.MultiGraph', 'igraph.Graph']
        raise NotImplementedError("Input graph must be one of: {}\nCurrently, type(graph) = {}".format("\n\n\t" + "\n\t".join(allowed), type(graph)))



"""
This module provides the VarLiNGAM class.

Classes:
    VarLiNGAM: class containing the VarLiNGAM causal discovery algorithm.
"""

from lingam.var_lingam import VARLiNGAM
import numpy as np
from causalflow.causal_discovery.CausalDiscoveryMethod import CausalDiscoveryMethod
from causalflow.graph.DAG import DAG
from causalflow.causal_discovery.baseline.pkgs import utils


class VarLiNGAM(CausalDiscoveryMethod):
    """VarLiNGAM causal discovery method."""
    
    def __init__(self, 
                 data, 
                 min_lag,
                 max_lag, 
                 verbosity, 
                 alpha = 0.05, 
                 resfolder = None,
                 neglect_only_autodep = False,
                 clean_cls = True):
        """
        Class constructor.

        Args:
            data (Data): data to analyse.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            verbosity (CPLevel): verbosity level.
            alpha (float, optional): PCMCI significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.
        """
        super().__init__(data, min_lag, max_lag, verbosity, alpha, resfolder, neglect_only_autodep, clean_cls)


    def run(self) -> DAG:
        """
        Run causal discovery algorithm.

        Returns:
            (DAG): estimated causal model.
        """
        split_by_causal_effect_sign = True

        model = VARLiNGAM(lags = self.max_lag, criterion='bic', prune=True)
        model.fit(self.data.d)

        m = model._adjacency_matrices
        am = np.concatenate([*m], axis=1)

        dag = np.abs(am) > self.alpha

        if split_by_causal_effect_sign:
            direction = np.array(np.where(dag))
            signs = np.zeros_like(dag).astype('int64')
            for i, j in direction.T:
                signs[i][j] = np.sign(am[i][j]).astype('int64')
            dag = signs

        dag = np.abs(dag)
        names = self.data.features
        res_dict = dict()
        for e in range(dag.shape[0]):
            res_dict[names[e]] = []
        for c in range(dag.shape[0]):
            for te in range(dag.shape[1]):
                if dag[c][te] == 1:
                    e = te%dag.shape[0]
                    t = te//dag.shape[0]
                    res_dict[names[e]].append((names[c], -t))
        self.CM = self._to_DAG(res_dict)
        
        if self.resfolder is not None: self.logger.close()
        return self.CM
    
    def _to_DAG(self, graph):
        """
        Re-elaborates the result in a DAG.
        
        Args:
            graph (dict): graph to convert into a DAG
            
        Returns:
            (DAG): result re-elaborated.
        """
        tmp_dag = DAG(self.data.features, self.min_lag, self.max_lag, self.neglect_only_autodep)
        tmp_dag.sys_context = dict()
        for t in graph.keys():
            for s in graph[t]:
                lag = abs(s[1])
                if lag >= self.min_lag and lag <= self.max_lag:
                    tmp_dag.add_source(t, s[0], utils.DSCORE, 0, s[1])
        return tmp_dag


#!/usr/bin/env python
"""
Implements the BaseArcDiagram and ArcDiagram classes.
"""

import numpy as np
import matplotlib.pyplot as plt

from ._graph_classes import BaseGraph, Graph
from ._node_layout import get_linear_layout, _get_preordered_linear_layout


class BaseArcDiagram(BaseGraph):
    """The arc diagram base class.

    Parameters
    ----------
    edges : list
        The edges of the graph, with each edge being represented by a (source node ID, target node ID) tuple.
    nodes : list or None, default None
        List of nodes. Required argument if any node in the graph is unconnected.
        If None, `nodes` is initialised to the set of the flattened `edges`.
    node_order : list or None, default None
        The ordering of nodes (left-to-right).
        If None, the node order is optimised such that the number of edge crossings is minimal.
    above : bool, default True
        If True, edges arc above the line of nodes.
        If False, edges arc below.
    node_layout : str or dict, default 'linear'
        If `node_layout` is a string, the node positions are computed using the indicated method:

        - 'linear'    : place nodes one a horizontal line

        If `node_layout` is a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_linear_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to :py:func:`get_arced_edge_paths`.
        Possible keyword arguments are:

        - :code:`rad` : float, default 1.
          Controls the curvature of the arcs.
        - :code:`selfloop_radius` : float, default 0.05 * np.linalg.norm(scale)
          Selfloop radius.
        - :code:`selfloop_angle` : float, default np.pi / 2
          Selfloop start angle

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`BaseGraph`, :py:class:`ArcDiagram`

    """

    def __init__(self, edges, nodes=None, node_layout='linear', node_order=None, above=True, *args, **kwargs):
        self.above = above
        kwargs.setdefault('node_layout_kwargs', dict())
        kwargs['node_layout_kwargs'].setdefault('node_order', node_order)
        if node_order:
            kwargs['node_layout_kwargs'].setdefault('reduce_edge_crossings', False)
        else:
            kwargs['node_layout_kwargs'].setdefault('reduce_edge_crossings', True)
        kwargs.setdefault("edge_layout_kwargs", dict())
        kwargs["edge_layout_kwargs"].setdefault("above", above)
        super().__init__(edges, nodes=nodes, node_layout=node_layout, edge_layout='arc', *args, **kwargs)


    def _get_node_positions(self, node_layout, node_layout_kwargs, edges):
        if len(self.nodes) == 1:
            return {self.nodes[0]: np.array([self.origin[0] + 0.5 * self.scale[0], self.origin[1] + 0.5 * self.scale[1]])}
        elif node_layout == 'linear':
            if node_layout_kwargs["node_order"]:
                node_order = node_layout_kwargs["node_order"]
                pad_by = node_layout_kwargs.get("pad_by", 0.05)
                return _get_preordered_linear_layout(
                    node_order, origin=self.origin, scale=self.scale, pad_by=pad_by)
            else:
                return get_linear_layout(
                    edges, nodes=self.nodes, origin=self.origin, scale=self.scale, **node_layout_kwargs)
        else:
            implemented = ['linear']
            msg = f"Node layout {node_layout} not implemented. Available layouts are:"
            for method in implemented:
                msg += f"\n\t{method}"
            raise NotImplementedError(msg)


class ArcDiagram(BaseArcDiagram, Graph):
    """
    Parses the given graph data object and initialises the :py:class:`BaseArcDiagram` object.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - networkx.Graph, igraph.Graph, or graph_tool.Graph object

    node_order : list or None, default None
        The ordering of nodes (left-to-right).
        If None, the node order is optimised such that the number of edge crossings is minimal.
    above : bool, default True
        If True, edges arc above the line of nodes.
        If False, edges arc below.
    node_layout : str or dict, default 'linear'
        If `node_layout` is a string, the node positions are computed using the indicated method:

        - 'linear'    : place nodes one a horizontal line

        If `node_layout` is a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_linear_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_cmap : matplotlib color map (default 'RdGy')
        Color map used to map edge weights to edge colors. Should be diverging.
        If edge weights are strictly positive, weights are mapped to the
        left hand side of the color map with vmin=0 and vmax=np.max(weights).
        If edge weights are positive and negative, then weights are mapped
        to colors such that a weight of zero corresponds to the center of the
        color map; the boundaries are set to +/- the maximum absolute weight.
        If the graph is unweighted or the edge colors are specified explicitly,
        this parameter is ignored.
    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to :py:func:`get_arced_edge_paths`.
        Possible keyword arguments are:

        - :code:`rad` : float, default 1.
          Controls the curvature of the arcs.
        - :code:`selfloop_radius` : float, default 0.05 * np.linalg.norm(scale)
          Selfloop radius.
        - :code:`selfloop_angle` : float, default np.pi / 2
          Selfloop start angle

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`BaseArcDiagram`, :py:class:`Graph`, :py:class:`InteractiveArcDiagram`

    """

    def __init__(self, graph, node_layout='linear', node_order=None, above=True, *args, **kwargs):
        self.above = above
        kwargs.setdefault('node_layout_kwargs', dict())
        kwargs['node_layout_kwargs'].setdefault('node_order', node_order)
        if node_order:
            kwargs['node_layout_kwargs'].setdefault('reduce_edge_crossings', False)
        else:
            kwargs['node_layout_kwargs'].setdefault('reduce_edge_crossings', True)
        kwargs.setdefault("edge_layout_kwargs", dict())
        kwargs["edge_layout_kwargs"].setdefault("above", above)
        Graph.__init__(self, graph, node_layout=node_layout, edge_layout='arc', *args, **kwargs)



"""
This module provides the TCDF class.

Classes:
    TCDF: class containing the TCDF causal discovery algorithm.
"""

from pathlib import Path
from subprocess import Popen, PIPE
import os
import json
from causalflow.graph.DAG import DAG
from causalflow.causal_discovery.CausalDiscoveryMethod import CausalDiscoveryMethod 
from causalflow.CPrinter import CP
from causalflow.causal_discovery.baseline.pkgs import utils


class TCDF(CausalDiscoveryMethod):
    """TCDF causal discovery method."""
    
    def __init__(self, 
                 data, 
                 min_lag,
                 max_lag, 
                 verbosity, 
                 resfolder = None,
                 neglect_only_autodep = False,
                 clean_cls = True):
        """
        Class constructor.

        Args:
            data (Data): data to analyse.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            verbosity (CPLevel): verbosity level.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.
        """
        super().__init__(data, min_lag, max_lag, verbosity, resfolder=resfolder, neglect_only_autodep=neglect_only_autodep, clean_cls=clean_cls)


    def run(self, 
            epochs=1000,  
            kernel_size=4, 
            dilation_coefficient=4, 
            hidden_layers=0, 
            learning_rate=0.01,
            cuda=False) -> DAG:
        """
        Run causal discovery algorithm.

        Returns:
            (DAG): estimated causal model.
        """
        # Remove all arguments from directory
        dir_path = os.path.dirname(os.path.realpath(__file__))
        Path(dir_path+"/args").mkdir(exist_ok=True)
        Path(dir_path+"/results").mkdir(exist_ok=True)
        script = dir_path + "/pkgs/TCDF-master/runTCDF" + ".py"
        r_arg_list = []
        r_arg_list.append("--epochs")
        r_arg_list.append(str(epochs))
        r_arg_list.append("--kernel_size")
        r_arg_list.append(str(kernel_size))
        r_arg_list.append("--dilation_coefficient")
        r_arg_list.append(str(dilation_coefficient))
        r_arg_list.append("--hidden_layers")
        r_arg_list.append(str(hidden_layers))
        r_arg_list.append("--learning_rate")
        r_arg_list.append(str(learning_rate))
        r_arg_list.append("--significance")
        r_arg_list.append(str(0.8))
        self.data.d.to_csv(dir_path + "/args/data.csv", index=False)
        r_arg_list.append("--data")
        r_arg_list.append(dir_path + "/args/data.csv")            
        if cuda: r_arg_list.append("--cuda")
        r_arg_list.append("--path")
        r_arg_list.append(dir_path)
        
        cmd = ["python", script] + r_arg_list
        p = Popen(cmd, cwd="./", stdin=PIPE, stdout=PIPE, stderr=PIPE)
        
        # Return R output or error
        output, error = p.communicate()
        CP.info(output.decode('utf-8'))
        if p.returncode == 0:
            g_dict = json.load(open(dir_path + "/results/tcdf_result.txt"))
            for key in g_dict.keys():
                key_list = []
                for elem in g_dict[key]:
                    key_list.append(tuple(elem))
                g_dict[key] = key_list
            utils.clean(dir_path)
            self.CM = self._to_DAG(g_dict)

            if self.resfolder is not None: self.logger.close()
            return self.CM
        else:
            utils.clean(dir_path)
            CP.warning('Python Error:\n {0}'.format(error))
            exit(0)
    
    
    def _to_DAG(self, graph):
        """
        Re-elaborate the result in a DAG.
        
        Args:
            graph (dict): graph to convert into a DAG

        Returns:
            (DAG): result re-elaborated.
        """
        tmp_dag = DAG(self.data.features, self.min_lag, self.max_lag, self.neglect_only_autodep)
        tmp_dag.sys_context = dict()
        for t in graph.keys():
            for s in graph[t]:
                lag = abs(s[1])
                if lag >= self.min_lag and lag <= self.max_lag:
                    tmp_dag.add_source(t, s[0], utils.DSCORE, 0, s[1])
        return tmp_dag


"""
This module provides various classes for the creation of random system of equations.

Classes:
    NoiseType: support class for handling different types of noise.
    PriorityOp: support class for handling different types of operator.
    RandomGraph: facilitate the creation of random graphs.
"""

import copy
from enum import Enum
import math
import random
from matplotlib import pyplot as plt
import numpy as np
from causalflow.graph.DAG import DAG
from causalflow.graph.PAG import PAG
from causalflow.preprocessing.data import Data

NO_CYCLES_THRESHOLD = 10


class NoiseType(Enum):
    """NoiseType Enumerator."""
    
    Uniform = 'uniform'
    Gaussian = 'gaussian'
    Weibull = 'weibull'
    
    
class PriorityOp(Enum):
    """PriorityOp Enumerator."""
    
    M = '*'
    D = '/'
    

class RandomGraph:
    """RandomGraph Class."""
    
    def __init__(self, nvars, nsamples, link_density, coeff_range: tuple, 
                 min_lag, max_lag, max_exp = None, noise_config: tuple = None, 
                 operators = ['+', '-', '*'], 
                 functions = ['','sin', 'cos', 'exp', 'abs', 'pow'],
                 n_hidden_confounders = 0,
                 n_confounded_vars = None):
        """
        Class constructor.

        Args:
            nvars (int): Number of variable.
            nsamples (int): Number of samples.
            link_density (int): Max number of parents per variable.
            coeff_range (tuple): Coefficient range. E.g. (-1, 1).
            min_lag (int): Min lagged dependency.
            max_lag (int): Max lagged dependency.
            max_exp (int): Max permitted exponent used by the 'pow' function. Used only if 'pow' is in the list of functions. Defaults to None.
            noise_config (tuple, optional): Noise configuration, e.g. (NoiseType.Uniform, -0.1, 0.1). Defaults to None.
            operators (list, optional): list of possible operators between variables. Defaults to ['+', '-', '*'].
            functions (list, optional): list of possible functions. Defaults to ['','sin', 'cos', 'exp', 'abs', 'pow'].
            n_hidden_confounders (int, optional): Number of hidden confounders. Defaults to 0.
            n_confounded_vars (int, optional): Number of confounded variables. If None, n_confounded_vars will be set as random.randint(2, nvars). Defaults to None.

        Raises:
            ValueError: max_exp cannot be None if functions list contains pow.
        """
        if 'pow' in functions and max_exp is None:
            raise ValueError('max_exp cannot be None if functions list contains pow')
               
        self.T = nsamples
        self.link_density = link_density
        self.coeff_range = coeff_range
        self.exponents = list(range(0, max_exp))
        self.min_lag = min_lag
        self.max_lag = max_lag
        self.n_hidden_confounders = n_hidden_confounders
        self.n_confounded = n_confounded_vars
        
        self.obsVar = ['X_' + str(i) for i in range(nvars)]
        self.hiddenVar = ['H_' + str(i) for i in range(n_hidden_confounders)]
        self.operators = operators
        self.functions = functions
        self.equations = {var: list() for var in self.obsVar + self.hiddenVar}
        self.confounders = {h: list() for h in self.hiddenVar}
        self.dependency_graph = {var: set() for var in self.obsVar + self.hiddenVar}
        self.PAG = None

        self.noise_config = noise_config
        self.noise = None
        if noise_config is not None:
            if noise_config[0] is NoiseType.Uniform:
                self.noise = np.random.uniform(noise_config[1], noise_config[2], (self.T, self.N))
            elif noise_config[0] is NoiseType.Gaussian:
                self.noise = np.random.normal(noise_config[1], noise_config[2], (self.T, self.N))
            elif noise_config[0] is NoiseType.Weibull:
                self.noise = np.random.weibull(noise_config[1], (self.T, self.N)) * noise_config[2]
    
    
    @property            
    def variables(self) -> list:
        """
        Retrieve the full set of observed and hidden variables.

        Returns:
            list: A list containing both observed and hidden variables.
        """
        return self.obsVar + self.hiddenVar 
    
    
    @property            
    def Nobs(self) -> int:
        """
        Return number of observable variables.

        Returns:
            int: number of observable variables.
        """
        return len(self.obsVar) 
    
         
    @property            
    def N(self) -> int:
        """
        Return total number of variables (observed and hidden).

        Returns:
            int: total number of variables.
        """
        return len(self.obsVar) + len(self.hiddenVar)
    
    
    @property
    def obsEquations(self) -> dict:
        """
        Return equations corresponding to the observed variables.

        Returns:
            dict: equations corresponding to the observed variables.
        """
        tmp = copy.deepcopy(self.equations)
        for h in self.hiddenVar: del tmp[h]
        return tmp
       
                
    def __build_equation(self, var_lagged_choice: list, var_contemp_choice: list, target_var) -> list:
        """
        Generate random equations.

        Args:
            var_lagged_choice (list): list of possible lagged parents for the target variable.
            var_contemp_choice (list): list of possible contemporaneous parents for the target variable.
            target_var (str): target variable.

        Returns:
            list: equation (list of tuple).
        """
        no_cycles_attempt = 0
        equation = []
        n_parents = random.randint(1, self.link_density)
        while len(equation) < n_parents:
            coefficient = random.uniform(self.coeff_range[0], self.coeff_range[1])
            lag = random.randint(self.min_lag, self.max_lag)
            if lag != 0:
                variable = random.choice(var_lagged_choice)
                var_lagged_choice.remove(variable)
            else:
                variable = random.choice(var_contemp_choice)
                var_contemp_choice.remove(variable)
                
            if not self.__creates_cycle((target_var, 0), (variable, lag)):
                operator = random.choice(self.operators)
                function = random.choice(self.functions)
                if function == 'pow':
                    exponent = random.choice(self.exponents)
                    term = (operator, coefficient, function, variable, lag, exponent)
                else:
                    term = (operator, coefficient, function, variable, lag)
                equation.append(term)
            else:
                no_cycles_attempt += 1
                if no_cycles_attempt >= NO_CYCLES_THRESHOLD:
                    raise ValueError("Cycle configuration impossible to be avoided!")
                
        return equation
    
    
    def __creates_cycle(self, target_var_lag, variable_lag) -> bool:
        """
        Check the presence of cycles.
        
        Specifically, it checks whether adding an edge from variable_lag to target_var_lag 
        creates a cycle considering only the same time lag

        Args:
            target_var_lag (str): target node.
            variable_lag (str): source node.

        Returns:
            bool: True if it finds cycles. Otherwise False.
        """
        target_var, target_lag = target_var_lag
        variable, lag = variable_lag

        visited = set()
        stack = [(variable, lag, [(variable, lag)], lag - target_lag)]
        while stack:
            current_var, current_lag, path, initial_lag_diff = stack.pop()
            if (current_var, current_lag) == target_var_lag:
                print(f"Cycle path: {' -> '.join([f'{var} (lag {l})' for var, l in [target_var_lag] + path])}")
                return True
            if (current_var, current_lag) not in visited:
                visited.add((current_var, current_lag))
                for neighbor_var, neighbor_lag in self.dependency_graph.get(current_var, []):
                    if (neighbor_var, neighbor_lag) not in visited:
                        # Check if the lag difference is the same as the initial lag difference
                        if (neighbor_lag - current_lag) == initial_lag_diff:
                            stack.append((neighbor_var, neighbor_lag, path + [(neighbor_var, neighbor_lag)], initial_lag_diff))
        # Update dependency graph
        self.dependency_graph[target_var].add(variable_lag)
        return False


    def gen_equations(self):
        """Generate random equations using the operator and function lists provided in the constructor."""
        for var in self.obsVar:
            var_lagged_choice = copy.deepcopy(self.obsVar)
            var_contemp_choice = copy.deepcopy(var_lagged_choice)
            var_contemp_choice.remove(var)
            self.equations[var] = self.__build_equation(var_lagged_choice, var_contemp_choice, var)
            
        for hid in self.hiddenVar:
            var_lagged_choice = copy.deepcopy(self.obsVar + self.hiddenVar)
            var_contemp_choice = copy.deepcopy(var_lagged_choice)
            var_contemp_choice.remove(hid)
            self.equations[hid] = self.__build_equation(var_lagged_choice, var_contemp_choice, hid)
            
        self.__add_conf_links()
               
        
    def __add_conf_links(self):
        """Add confounder links to a predefined causal model."""
        no_cycles_attempt = 0
        self.expected_bidirected_links = list()
        firstvar_choice = copy.deepcopy(self.obsVar)
        for hid in self.hiddenVar:
            tmp_n_confounded = 0
            isContemporaneous = random.choice([True, False])
            n_confounded = random.randint(2, self.Nobs) if self.n_confounded is None else self.n_confounded 
            
            if isContemporaneous:
                lag = random.randint(self.min_lag, self.max_lag)
                confVar = list()
                while tmp_n_confounded < n_confounded:
                    variable = random.choice(firstvar_choice)
                    
                    if not self.__creates_cycle((variable, 0), (hid, lag)):
                        tmp_n_confounded += 1
                        firstvar_choice.remove(variable)
                        confVar.append(variable)
                                                                    
                        function = random.choice(self.functions)
                        coefficient = random.uniform(self.coeff_range[0], self.coeff_range[1])
                        operator = random.choice(self.operators)
                        if function == 'pow':
                            exponent = random.choice(self.exponents)
                            term = (operator, coefficient, function, hid, lag, exponent)
                        else:
                            term = (operator, coefficient, function, hid, lag)
                            
                        #! NOTE: This is to remove the true link between confounded variable for ensuring 
                        #! that the link due to the confounder is classified as spurious
                        if len(confVar) > 1:
                            for source in confVar:
                                tmp = copy.deepcopy(confVar)
                                tmp.remove(source)
                                for target in tmp:
                                    if (source, 0) in self.get_Adj()[target]:
                                        self.equations[target] = list(filter(lambda item: item[3] != source and item[3] != 0, self.equations[target]))
                        self.equations[variable].append(term)
                    
                        self.confounders[hid].append((variable, lag))
                    else:
                        no_cycles_attempt += 1
                        if no_cycles_attempt >= NO_CYCLES_THRESHOLD:
                            raise ValueError("Impossible to avoid the cycle configuration!")
                for source in confVar:
                    tmp = copy.deepcopy(confVar)
                    tmp.remove(source)
                    for target in tmp:
                        if not (source, 0) in self.get_Adj()[target]:
                            self.expected_bidirected_links.append({target: (source, 0)})
            else:    
                var_choice = copy.deepcopy(self.obsVar)
                firstConf = True
                source = None
                sourceLag = None
                targets = list()
                
                while tmp_n_confounded < n_confounded:
                    if firstConf:
                        variable = random.choice(firstvar_choice)
                        lag = random.randint(self.min_lag, self.max_lag - 1)
                        sourceLag = lag
                    else:
                        variable = random.choice(var_choice)
                        lag = random.randint(sourceLag + 1, self.max_lag)
                     
                    if not self.__creates_cycle((variable, 0), (hid, lag)):
                        tmp_n_confounded += 1
                        var_choice.remove(variable)
                        if firstConf:
                            firstvar_choice.remove(variable)
                            firstConf = False
                            source = variable
                        else:
                            targets.append((variable, lag))
                            
                            #! NOTE: This is to remove the true link between confounded variable for ensuring 
                            #! that the link due to the confounder is classified as spurious
                            if (source, lag - sourceLag) in self.get_Adj()[variable]:
                                self.equations[variable] = list(filter(lambda item: item[3] != source and item[3] != lag - sourceLag, self.equations[variable]))
                        
                        function = random.choice(self.functions)
                        coefficient = random.uniform(self.coeff_range[0], self.coeff_range[1])
                        operator = random.choice(self.operators)
                        if function == 'pow':
                            exponent = random.choice(self.exponents)
                            term = (operator, coefficient, function, hid, lag, exponent)
                        else:
                            term = (operator, coefficient, function, hid, lag)
                        self.equations[variable].append(term)
                    
                        self.confounders[hid].append((variable, lag))
                    else:
                        no_cycles_attempt += 1
                        if no_cycles_attempt >= NO_CYCLES_THRESHOLD:
                            raise ValueError("Cycle configuration impossible to be avoided!")
                        
                # Lagged bidirected links
                for v in targets:
                    if not (source, v[1] - sourceLag) in self.get_Adj()[v[0]]:
                        self.expected_bidirected_links.append({v[0]: (source, v[1] - sourceLag)})
                # Contemporaneous bidirected links
                for source in targets:
                    tmp = copy.deepcopy(targets)
                    tmp.remove(source)
                    for target in tmp:
                        if not (source, 0) in self.get_Adj()[target[0]]:
                            self.expected_bidirected_links.append({target[0]: (source[0], 0)})


    def print_equations(self):
        """Print the generated equations."""
        toprint = list()
        for target, eq in self.equations.items():
            equation_str = target + '(t) = '
            for i, term in enumerate(eq):
                if len(term) == 6:
                    operator, coefficient, function, variable, lag, exponent = term
                    coefficient = round(coefficient, 2)
                    if i != 0: 
                        term_str = f"{operator} {coefficient} * {function}({variable}, {exponent})(t-{lag}) "
                    else:
                        term_str = f"{coefficient} * {function}({variable}, {exponent})(t-{lag}) "
                else:
                    operator, coefficient, function, variable, lag = term
                    coefficient = round(coefficient, 2)
                    if function != '':
                        if i != 0: 
                            term_str = f"{operator} {coefficient} * {function}({variable})(t-{lag}) "
                        else:
                            term_str = f"{coefficient} * {function}({variable})(t-{lag}) "
                    else:
                        if i != 0: 
                            term_str = f"{operator} {coefficient} * {variable}(t-{lag}) "
                        else:
                            term_str = f"{coefficient} * {variable}(t-{lag}) "
                        
                equation_str += term_str
            toprint.append(equation_str)
        eq = "\n".join(toprint)
        print(eq)
        return eq
            
            
    def __evaluate_term(self, term, t, data) -> tuple:
        """
        Evaluate single term componing an equation.

        Args:
            term (tuple): term to evaluate.
            t (int): time step.
            data (numpy array): time-series.

        Returns:
            tuple: operator and value of the term.
        """
        operator, coefficient, function, variable, *args = term
        if function == '':
            lag = args[0]
            term_value = coefficient * (data[t - lag, self.variables.index(variable)])
        elif function == 'pow':
            lag, exponent = args
            term_value = coefficient * data[t - lag, self.variables.index(variable)] ** exponent
        elif function == 'abs':
            lag = args[0]
            term_value = coefficient * abs(data[t - lag, self.variables.index(variable)])
        else:
            lag = args[0]
            term_value = coefficient * getattr(math, function)(data[t - lag, self.variables.index(variable)])
        return operator, term_value
    
    
    def __handle_priority_operator(self, eq) -> list:
        """
        Evaluate all the terms with operato * ans /.

        Args:
            eq (list): equation (list of term).

        Returns:
            list: equation with all * and / evaluated.
        """
        op = '*'
        while (op in eq):
            op_i = eq.index(op)
            op1_i = op_i - 1
            op2_i = op_i + 1
            eq[op1_i] = eq[op1_i] * eq[op2_i]
                
            indices_set = set([op_i, op2_i])
            eq = [item for i, item in enumerate(eq) if i not in indices_set]
            
        op = '/'
        while (op in eq):
            op_i = eq.index(op)
            op1_i = op_i - 1
            op2_i = op_i + 1
            eq[op1_i] = eq[op1_i] / eq[op2_i]
                
            indices_set = set([op_i, op2_i])
            eq = [item for i, item in enumerate(eq) if i not in indices_set]
            
        return eq
     

    def __evaluate_equation(self, equation, t, data) -> float:
        """
        Evaluate equation.

        Args:
            equation (list): equation (list of term).
            t (int): time step.
            data (numpy array): time-series.
        
        Returns:
            float: equation value.
        """
        eq = list()
        for i, term in enumerate(equation):
            operator, term = self.__evaluate_term(term, t, data)
            if i == 0:
                eq.append(term)
            else:
                eq.append(operator)
                eq.append(term)
                
        # Handle * and / before + and -
        eq = self.__handle_priority_operator(eq)
        
        equation_value = eq.pop(0)
        for i in range(0, len(eq), 2):
            op = eq[i]
            term = eq[i+1]
            if op == '+': equation_value = equation_value + term
            elif op == '-': equation_value = equation_value - term
        return equation_value


    def gen_obs_ts(self) -> tuple:
        """
        Generate time-series data.

        Returns:
            tuple: (Data obj with hidden vars, Data obj without hidden vars).
        """
        np_data = np.zeros((self.T, self.N))
        for t in range(self.T):
            if t < self.max_lag:
                for target, eq in self.equations.items():
                    np_data[t, self.variables.index(target)] = self.noise[t, self.variables.index(target)]
            else:
                for target, eq in self.equations.items():
                    np_data[t, self.variables.index(target)] = self.__evaluate_equation(eq, t, np_data)
                    if self.noise is not None: np_data[t, self.variables.index(target)] += self.noise[t, self.variables.index(target)]
                    
        data = Data(np_data, self.variables)
        only_obs = copy.deepcopy(data)
        only_obs.shrink(self.obsVar)
        return data, only_obs
    
    
    def gen_interv_ts(self, interventions, obs) -> dict:
        """
        Generate time-series corresponding to intervention(s).

        Args:
            interventions (dict): dictionary {INT_VAR : {INT_LEN: int_len, INT_VAL: int_val}}.
            obs (DataFrame): Observational DataFrame.
        
        Returns:
            dict: {interventional variable: interventional time-series data}.
        """
        starting_point = obs.values
        int_data = dict()
        for int_var in interventions:
            T = int(interventions[int_var]["T"])
            if self.noise_config is not None:
                if self.noise_config[0] is NoiseType.Uniform:
                    int_noise = np.random.uniform(self.noise_config[1], self.noise_config[2], (T, self.N))
                elif self.noise_config[0] is NoiseType.Gaussian:
                    int_noise = np.random.normal(self.noise_config[1], self.noise_config[2], (T, self.N))
                elif self.noise_config[0] is NoiseType.Weibull:
                    int_noise= np.random.weibull(self.noise_config[1], (self.T, self.N)) * self.noise_config[2]
            np_data = np.zeros((T, self.N))
            np_data[0:self.max_lag, :] = starting_point[len(starting_point)-self.max_lag:,:]

            for t in range(self.max_lag, T):
                for target, eq in self.equations.items():
                    if target != int_var:
                        np_data[t, self.variables.index(target)] = self.__evaluate_equation(eq, t, np_data)
                        if self.noise_config is not None: np_data[t, self.variables.index(target)] += int_noise[t, self.variables.index(target)]
                    else:
                        np_data[t, self.variables.index(target)] = interventions[int_var]["VAL"]
                        
            int_data[int_var] = Data(np_data, self.variables)
            int_data[int_var].shrink(self.obsVar)
            starting_point = np_data
        return int_data
    
    
    def get_DPAG(self) -> dict:
        """
        Output the PAG starting from a DAG.

        Returns:
            dict: scm.
        """
        if self.PAG is None:
            scm = self.get_Adj(withHidden=True)
            self.PAG = PAG(scm, self.max_lag, self.hiddenVar)
        return self.PAG.convert2Graph()
    
    
    def get_Adj(self, withHidden = False) -> dict:
        """
        Output the Structural Causal Model.

        Args:
            withHidden (bool, optional): include hidden variables. Default to False.
            
        Returns:
            dict: scm.
        """
        eqs = self.equations if withHidden else self.obsEquations
        scm = {target : list() for target in eqs.keys()}
        for target, eq in eqs.items():
            for term in eq:
                if len(term) == 6:
                    _, _, _, variable, lag, _ = term
                else:
                    _, _, _, variable, lag = term
                if variable not in scm.keys(): continue # NOTE: this is needed to avoid adding hidden vars
                scm[target].append((variable, -abs(lag)))
        return scm
    
    
    def print_SCM(self, withHidden = False):
        """
        Print the Structural Causal Model.
        
        Args:
            withHidden (bool, optional): include hidden variables. Default to False.
        """
        scm = self.get_Adj(withHidden)
        for t in scm: print(t + ' : ' + str(scm[t]))    
          
        
    def intervene(self, int_var, int_len, int_value, obs) -> dict:
        """
        Generate intervention on a single variable.

        Args:
            int_var (str): variable name.
            int_len (int): intervention length.
            int_value (float): intervention value.
            obs (DataFrame): Observational DataFrame.
            
        Returns:
            dict: {interventional variable: interventional time-series data}.
        """
        if not isinstance(int_var, list): int_var = [int_var]
        if not isinstance(int_len, list): int_len = [int_len]
        if not isinstance(int_value, list): int_value = [int_value]
        return self.gen_interv_ts({v: {"T": l, "VAL": val} for v, l, val in zip(int_var, int_len, int_value)}, obs)
    
    
    def ts_dag(self, withHidden = False, save_name = None, randomColors = False):
        """
        Draw a Time-seris DAG.

        Args:
            withHidden (bool, optional): bit to decide whether to output the SCM including the hidden variables or not. Defaults to False.
            save_name (str, optional): figure path. Defaults to None.
            randomColors (bool, optional): random color for each node. Defaults to False.
        """
        gt = self.get_Adj(withHidden) if withHidden else self.get_DPAG()
        var = self.variables if withHidden else self.obsVar
        g = DAG(var, self.min_lag, self.max_lag, False, gt)
        node_color = []

        tab_colors = plt.cm.get_cmap('tab20', 20).colors  # You can adjust the number of colors if needed
        avail_tab_colors = list(copy.deepcopy(tab_colors))
        for t in g.g:
            if t in self.hiddenVar:
                node_color.append('peachpuff')
            else:
                if randomColors :
                    c = random.randint(0, len(avail_tab_colors)-1)
                    node_color.append(avail_tab_colors[c])
                    avail_tab_colors.pop(c)
                else:
                    node_color.append('orange')
                    
        # Edges color definition
        edge_color = dict()
        for t in g.g:
            for s in g.g[t].sources:
                s_index = len(g.g.keys())-1 - list(g.g.keys()).index(s[0])
                t_index = len(g.g.keys())-1 - list(g.g.keys()).index(t)
                
                s_lag = self.max_lag - s[1]
                t_lag = self.max_lag
                while s_lag >= 0:
                    s_node = (s_lag, s_index)
                    t_node = (t_lag, t_index)
                    if s[0] in self.hiddenVar:
                        edge_color[(s_node, t_node)] = 'gainsboro'
                    else:
                        edge_color[(s_node, t_node)] = 'gray'
                        
                    s_lag -= 1
                    t_lag -= 1
                        
        g.ts_dag(save_name = save_name, node_color = node_color, edge_color = edge_color, min_cross_width=2, max_cross_width=5, x_disp=1, node_size=6)




"""
This module provides the WSDynamic class.

Classes:
    WSDynamic: Subsampling method with dynamic window size based on entropy analysis.
"""

import ruptures as rpt
from causalflow.preprocessing.subsampling_methods.EntropyBasedMethod import EntropyBasedMethod
from causalflow.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod, SSMode


class WSDynamic(SubsamplingMethod, EntropyBasedMethod):
    """Subsampling method with dynamic window size based on entropy analysis."""
    
    def __init__(self, window_min_size, entropy_threshold):
        """
        Class constructor.

        Args:
            window_min_size (int): minimun window size.
            entropy_threshold (float): entropy threshold.

        Raises:
            ValueError: if window_min_size == None.
        """
        SubsamplingMethod.__init__(self, SSMode.WSDynamic)
        EntropyBasedMethod.__init__(self, entropy_threshold)
        if window_min_size is None:
            raise ValueError("window_type = DYNAMIC but window_min_size not specified")
        self.wms = window_min_size
        self.ws = None


    def dataset_segmentation(self):
        """Segment dataset based on breakpoint analysis and a min window size."""
        de = self.create_rounded_copy()
        algo = rpt.Pelt(model = "l2", min_size = self.wms).fit(de)
        seg_res = algo.predict(pen = 10)
        self.segments = [(seg_res[i - 1], seg_res[i]) for i in range(1, len(seg_res))]
        self.segments.insert(0, (0, seg_res[0]))


    def run(self):
        """
        Run subsampler.

        Returns:
            (list[int]): indexes of the remaining samples.
        """
        # build list of segment
        self.dataset_segmentation()

        # compute entropy moving window
        self.moving_window_analysis()

        # extracting subsampling procedure results
        idxs = self.extract_indexes()

        return idxs


"""
This module provides the CAnDOIT class.

Classes:
    CAnDOIT: class containing the CAnDOIT causal discovery algorithm.
"""

import copy
import pickle
import numpy as np
import pandas as pd
from causalflow.causal_discovery.tigramite.independence_tests.independence_tests_base import CondIndTest
from causalflow.selection_methods.SelectionMethod import SelectionMethod
from causalflow.CPrinter import CPLevel, CP
from causalflow.basics.constants import *
from causalflow.basics.utils import *
from causalflow.preprocessing.data import Data 
from causalflow.causal_discovery.CausalDiscoveryMethod import CausalDiscoveryMethod 
from causalflow.causal_discovery.support.myLPCMCI import myLPCMCI
from causalflow.graph.DAG import DAG

class CAnDOIT(CausalDiscoveryMethod):
    """CAnDOIT causal discovery method."""

    def __init__(self, 
                 observation_data: Data, 
                 intervention_data: dict, 
                 min_lag, max_lag,
                 sel_method: SelectionMethod, val_condtest: CondIndTest, 
                 verbosity: CPLevel, 
                 f_alpha = 0.05, 
                 alpha = 0.05, 
                 resfolder = None,
                 neglect_only_autodep = False,
                 exclude_context = True,
                 plot_data = False,
                 clean_cls = True):
        """
        Class contructor.

        Args:
            observation_data (Data): observational data to analyse.
            intervention_data (dict): interventional data to analyse in the form {INTERVENTION_VARIABLE : Data (same variables of observation_data)}.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            sel_method (SelectionMethod): selection method.
            val_condtest (CondIndTest): validation method.
            verbosity (CPLevel): verbosity level.
            f_alpha (float, optional): filter significance level. Defaults to 0.05.
            alpha (float, optional): PCMCI significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            exclude_context (bool, optional): Bit for neglecting context variables. Defaults to False.
            plot_data (bool, optional): Bit for plotting your data. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.
        """
        self.obs_data = observation_data
        self.systems = observation_data.features
        self.contexts = []
        self.sys_context = {}
        for k in intervention_data.keys():
            self.contexts.append("C" + k)
            self.sys_context[k] = "C" + k
        self.vars = self.systems + self.contexts
        
        self.f_alpha = f_alpha
        self.sel_method = sel_method
        self.val_condtest = val_condtest
        self.exclude_context = exclude_context
        super().__init__(self.obs_data, min_lag, max_lag, verbosity, alpha, resfolder, neglect_only_autodep, clean_cls)
        
        # Create filter and validator data
        self.filter_data, self.validator_data = self._prepare_data(self.obs_data, intervention_data, plot_data)
        
        
        CP.info("\n")
        CP.info(DASH)
        CP.info("Observational data length: " + str(observation_data.T))
        CP.info("Interventional data length: " + str(sum([d.T for d in intervention_data.values()])))
        CP.info("Min lag time: " + str(min_lag))
        CP.info("Max lag time: " + str(max_lag))
        CP.info("Filter significance level: " + str(f_alpha))
        CP.info("PCMCI significance level: " + str(alpha))
        CP.info("Selection method: " + sel_method.name)
        
            
    @property    
    def isThereInterv(self) -> bool:
        """
        Check whether an intervention is present or not.

        Returns:
            bool: flag to identify if an intervention is present or not.
        """
        return len(list(self.sys_context.keys())) > 0
    
    
    def JCI_assumptions(self):
        """Initialise the algorithm initial causal structure with the JCI assumptions."""
        # ! JCI Assmpution 1: No system variable causes any context variable
        # ! JCI Assmpution 2: No context variable is confounded with a system variable
        # ! JCI Assmpution 3: The context distribution contains no (conditional) independences

        knowledge = {self.vars.index(f): dict() for f in self.vars}
        
        # ! JCI Assmpution 1
        for k in self.contexts:
            for x in self.systems:
                for tau_i in range(0, self.max_lag + 1):
                    knowledge[self.vars.index(k)][(self.vars.index(x), -tau_i)] = ''
            
        # ! JCI Assmpution 2
        for k in self.contexts:
            for x in self.systems:
                if x not in self.sys_context or (x in self.sys_context and k != self.sys_context[x]):
                    for tau_i in range(0, self.max_lag + 1): knowledge[self.vars.index(x)][(self.vars.index(k), -tau_i)] = ''
                elif x in self.sys_context and k == self.sys_context[x]:
                    knowledge[self.vars.index(x)][(self.vars.index(k), 0)] = '-->'
                    knowledge[self.vars.index(k)][(self.vars.index(x), 0)] = '<--'
                    for tau_i in range(1, self.max_lag + 1): knowledge[self.vars.index(x)][(self.vars.index(k), -tau_i)] = ''
                    for tau_i in range(1, self.max_lag + 1): knowledge[self.vars.index(k)][(self.vars.index(x), -tau_i)] = ''
        
        # ! JCI Assmpution 3
        for k1 in self.contexts:
            for k2 in remove_from_list(self.contexts, k1):
                knowledge[self.vars.index(k1)][(self.vars.index(k2), 0)] = '<->'
                # for tau_i in range(0, self.max_lag + 1): knowledge[self.vars.index(k1)][(self.vars.index(k2), -tau_i)] = '<->'
        
        # ! This models the context variables as chain across different time steps
        for k in self.contexts:
            for tau_i in range(1, self.max_lag + 1):
                knowledge[self.vars.index(k)][(self.vars.index(k), -tau_i)] = '-->' if tau_i == 1 else ''
        
                          
                
                                              
        out = {}
        for j in range(len(self.vars)):
            inner_dict = {} 
            
            for i in range(len(self.vars)):
                for tau_i in range(0, self.max_lag + 1):
                    if tau_i > 0 or i != j:
                        value = "o?>" if tau_i > 0 else "o?o"
                        inner_dict[(i, -tau_i)] = value
                           
            out[j] = inner_dict

        for j, links_j in knowledge.items():
            for (i, lag_i), link_ij in links_j.items():
                if link_ij == "":
                    del out[j][(i, lag_i)]
                else: 
                    out[j][(i, lag_i)] = link_ij
        return out
    

    def run_filter(self):
        """Run filter method."""
        CP.info("Selecting relevant features among: " + str(self.filter_data.features))
       
        self.sel_method.initialise(self.obs_data, self.f_alpha, self.min_lag, self.max_lag, self.CM)
        self.CM = self.sel_method.compute_dependencies()
        
    
    def run_validator(self, link_assumptions = None) -> DAG:
        """
        Run Validator (LPCMCI).

        Args:
            link_assumptions (dict, optional): link assumption with context. Defaults to None.

        Returns:
            DAG: causal model with context.
        """
        self.validator = myLPCMCI(self.validator_data,
                                self.min_lag, self.max_lag,
                                self.sys_context,
                                self.val_condtest,
                                CP.verbosity,
                                self.alpha)
        causal_model = self.validator.run(link_assumptions)
        causal_model.sys_context = self.CM.sys_context      
 
        return causal_model
      
    
    def run(self, remove_unneeded = True, nofilter = True) -> DAG:
        """
        Run CAnDOIT.
        
        Args:
            remove_unneeded (bool, optional): Bit to remove unneeded (isolated) variables. Defaults to True.
            nofilter (bool, optional): Bit to run CAnDOIT without filter. Defaults to False.
            
        Returns:
            DAG: causal model.
        """
        link_assumptions = None
            
        if not nofilter:
            #FIXME: to include also the filter. for now this is wrong
            ## 1. FILTER
            self.run_filter()
        
            # list of selected features based on filter dependencies
            self.CM.remove_unneeded_features()
            if not self.CM.features: return None, None
            
            self.obs_data.shrink(self.CM.features)
            f_dag = copy.deepcopy(self.CM)
        
            ## 2. VALIDATOR
            # Add dependencies corresponding to the context variables 
            # ONLY if the the related system variable is still present
            self.CM.add_context() 

            # shrink dataframe d by using the filter result
            self.validator_data.shrink(self.CM.features)
                    
            # selected links to check by the validator
            link_assumptions = self.CM.get_link_assumptions()
                
        else:
            # fullg = DAG(self.validator_data.features, self.min_lag, self.max_lag, False)
            # fullg.sys_context = self.CM.sys_context
            link_assumptions = self.JCI_assumptions()
            
        # calculate dependencies on selected links
        self.CM = self.run_validator(link_assumptions)
               
        # list of selected features based on validator dependencies
        if remove_unneeded: self.CM.remove_unneeded_features()
        if self.exclude_context: self.CM.remove_context()

        self.save()
            
        return self.CM
            
    
    def load(self, res_path):
        """
        Load previously estimated result.

        Args:
            res_path (str): pickle file path.
        """
        with open(res_path, 'rb') as f:
            r = pickle.load(f)
            self.CM = r['causal_model']
            self.f_alpha = r['filter_alpha']
            self.alpha = r['alpha']
            self.dag_path = r['dag_path']
            self.ts_dag_path = r['ts_dag_path']
            
            
    def save(self):
        """Save causal discovery result as pickle file if resfolder is set."""
        if self.respath is not None:
            if self.CM:
                res = dict()
                res['causal_model'] = copy.deepcopy(self.CM)
                res['features'] = copy.deepcopy(self.CM.features)
                res['filter_alpha'] = self.f_alpha
                res['alpha'] = self.alpha
                res['dag_path'] = self.dag_path
                res['ts_dag_path'] = self.ts_dag_path
                with open(self.respath, 'wb') as resfile:
                    pickle.dump(res, resfile)
            else:
                CP.warning("Causal model impossible to save")
                
                
    def _prepare_data(self, obser_data, inter_data, plot_data):
        """
        Prepare data for filter and validator phases.
        
        Args:
            obser_data (Data): observational data.
            inter_data (Data): interventional data.
            plot_data (bool): boolean bit to plot the generated data.

        Returns:
            Data, Data: filter data obj and validator data obj.
        """
        # Filter phase data preparation
        filter_data = copy.deepcopy(obser_data.d)
        for int_data in inter_data.values(): filter_data = pd.concat([filter_data, int_data.d], axis = 0, ignore_index = True)
        filter_data = Data(filter_data, vars = obser_data.features)
        
        # Validator phase data preparation
        validator_data = copy.deepcopy(obser_data.d)
        context_vars = dict()
        for int_var, int_data in inter_data.items():
            
            # Create context variable name
            context_varname = 'C' + int_var
            
            # Store a dict of context variable and system variable corresponding to an intervention
            self.CM.sys_context[int_var] = context_varname
            
            # Create context variable data
            # context_data = np.ones(shape=int_data.d[int_var].shape)
            context_data = int_data.d[int_var]
            context_start = len(validator_data)
            context_end = context_start + len(context_data)
            context_vars[context_varname] = {'data': context_data, 'start': context_start, 'end': context_end}
            
            validator_data = pd.concat([validator_data, int_data.d], axis = 0, ignore_index = True)
            
        for var in context_vars:
            new_column = np.zeros(shape = (len(validator_data),))
            new_column[context_vars[var]['start']: context_vars[var]['end']] = context_vars[var]['data']
            validator_data[var] = new_column
        
        validator_data = Data(validator_data, vars = list(validator_data.columns))
        
        if plot_data: validator_data.plot_timeseries()
        return filter_data, validator_data


"""
This module provides the support classes and various constants.

Classes:
    LabelType: label type.
    LinkType: link type.
    ImageExt: image extention.
"""

from enum import Enum

SOURCE = 'source'
SCORE = 'score'
PVAL = 'pval'
LAG = 'lag'
TYPE = 'type'

DASH = '-' * 55
SEP = "/"
RES_FILENAME = "res.pkl"
DAG_FILENAME = "dag"
TSDAG_FILENAME = "ts_dag"
LOG_FILENAME = "log.txt"


class LabelType(Enum):
    """LabelType Enumerator."""
    
    Lag = "Lag"
    Score = "Score"
    NoLabels = "NoLabels"
    
    
class DataType(Enum):
    """DataType Enumerator."""
    
    Continuos = 0
    Discrete = 1
    
    
class LinkType(Enum):
    """LinkType Enumerator."""
    
    Directed = "-->"
    Uncertain = "o-o"
    Bidirected = "<->"
    HalfUncertain = "o->"
    
    
class ImageExt(Enum):
    """ImageExt Enumerator."""
    
    PNG = ".png"
    PDF = ".pdf"
    JPG = ".jpg"